#pragma once

#include <petsc/private/petscimpl.h>

#include <petsc/private/cpp/macros.hpp>
#include <petsc/private/cpp/utility.hpp>
#include <petsc/private/cpp/type_traits.hpp>
#include <petsc/private/cpp/tuple.hpp>
#include <petsc/private/cpp/array.hpp>
#include <petsc/private/cpp/functional.hpp>
#include <petsc/private/cpp/register_finalize.hpp>
#include <petsc/private/cpp/crtp.hpp>
