// the configured options and settings for Tutorial
// Copyright (c) 2019 University of Oregon
// Distributed under the BSD Software License
// (See accompanying file LICENSE.txt)
#define PerfStubs_VERSION_MAJOR 0
#define PerfStubs_VERSION_MINOR 1
// #define PERFSTUBS_USE_TIMERS
#define PERFSTUBS_USE_DEFAULT_IMPLEMENTATION
/* #undef PERFSTUBS_USE_STATIC */
