/*
    This include file allows you to use ANY public PETSc function
*/
#pragma once

#include <petscbag.h>
#include <petsctime.h>
#include <petscbt.h>
#include <petscmatlab.h>
#include <petscdraw.h>
#include <petscdevice.h>
#include <petscbm.h>

#include <petscsf.h>
#include <petscsection.h>
#include <petscdmda.h>
#include <petscdmcomposite.h>
#include <petscdmpatch.h>
#include <petscdmplex.h>
#include <petscdmplextransform.h>
#include <petscdmredundant.h>
#include <petscdmshell.h>
#include <petscdmsliced.h>
#include <petscdmswarm.h>
#include <petscdmstag.h>
#include <petscdmforest.h>
#include <petscdmnetwork.h>
#include <petscdmadaptor.h>

#include <petscdmlabelephemeral.h>

#include <petsclandau.h>

#include <petscdt.h>
#include <petscds.h>

#include <petsccharacteristic.h>

#include <petscts.h>
#include <petsctao.h>
