#pragma once

/* MANSEC = DM */
/* SUBMANSEC = DMPlex */

/*S
  DMPlexTransform - An object encapsulating a `DMPLEX` transformation

  Level: intermediate

.seealso: [](ch_dmbase), `DMPLEX`, `DMPlexTransformCreate()`, `DMPlexTransformType`
S*/
typedef struct _p_DMPlexTransform *DMPlexTransform;
