% Sample calling syntax for testing taopounders and comparing to fminsearch
% TestingFinalize is called after solving all the instances

% Store the results matrix to a file called results.mat
save results Results;

