double precision p1, p2, p3, p4, r1, r2, r3, r4 double precision e1, e2, e3, e4, v1, v2, v3, v4 double precision ru1, ru2, ru3, ru4, et1, et2, et3, et4 double precision csubv, gamma, a1, a2, a3, a4,
  sspd

      common /
    passit / p1,
  p2, p3, p4, r1, r2, r3, r4, e1, e2, e3, e4, v1, v2, v3, v4, ru1, ru2, ru3, ru4, et1, et2, et3, et4, csubv, gamma, a1, a2, a3, a4, sspd
